<?php

namespace Drupal\sfc\Plugin\SingleFileComponent;

use Drupal\Core\Form\FormStateInterface;
use Drupal\sfc\ComponentBase;

/**
 * Contains a component that renders based on the contents of a *.sfc file.
 *
 * @SingleFileComponent(
 *   id = "sfc_simple_component",
 *   deriver = "\Drupal\sfc\Plugin\Derivative\SimpleComponentDeriver",
 * )
 */
class SimpleComponent extends ComponentBase {

  /**
   * An array of file data.
   *
   * @var array
   */
  protected $fileData;

  /**
   * Loads, parses, and returns simple file data.
   *
   * @return array
   *   An array containing keys that match up with ComponentBase's constants.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   If parsing the file failed.
   */
  protected function getFileData() {
    if (!empty($this->fileData)) {
      return $this->fileData;
    }
    $filename = $this->pluginDefinition['simple_file'];
    ob_start();
    require $filename;
    $content = '<root>' . ob_get_contents() . '</root>';
    ob_end_clean();
    $new_data = [
      'JS' => NULL,
      'ATTACH' => NULL,
      'DETACH' => NULL,
      'CSS' => NULL,
      'TEMPLATE' => NULL,
      'SELECTOR' => NULL,
      'DEPENDENCIES' => NULL,
      'prepareContext' => NULL,
      'buildContextForm' => NULL,
      'validateContextForm' => NULL,
      'submitContextForm' => NULL,
    ];
    $document = new \DOMDocument();
    @$document->loadHTML($content);
    $xpath = new \DOMXPath($document);
    /** @var \DOMElement $node */
    foreach ($xpath->query('//root/script') as $node) {
      $type = $node->getAttribute('data-type');
      if (!$type) {
        $new_data['JS'] = $this->getInnerHtml($node);
      }
      elseif ($type === 'attach') {
        $new_data['ATTACH'] = $this->getInnerHtml($node);
      }
      elseif ($type === 'detach') {
        $new_data['DETACH'] = $this->getInnerHtml($node);
      }
    }
    foreach ($xpath->query('//root/style') as $node) {
      $new_data['CSS'] = $this->getInnerHtml($node);
    }
    foreach ($xpath->query('//root/template') as $node) {
      $new_data['TEMPLATE'] = trim($this->getInnerHtml($node), "\n");
    }
    if (empty($new_data['TEMPLATE'])) {
      $new_data['TEMPLATE'] = '';
    }
    if (isset($selector)) {
      $new_data['SELECTOR'] = $selector;
    }
    if (isset($dependencies)) {
      $new_data['DEPENDENCIES'] = $dependencies;
    }
    $callbacks = [
      'prepareContext',
      'buildContextForm',
      'validateContextForm',
      'submitContextForm',
    ];
    foreach ($callbacks as $callback) {
      if (isset($$callback) && is_callable($$callback)) {
        $new_data[$callback] = $$callback;
      }
    }
    $this->fileData = $new_data;
    return $this->fileData;
  }

  /**
   * {@inheritdoc}
   */
  public function prepareContext(array &$context) {
    $data = $this->getFileData();
    if ($data['prepareContext']) {
      $data['prepareContext']($context);
    }
  }

  /**
   * Gets the inner HTML for a DOM node.
   *
   * @param \DOMNode $node
   *   The DOM node.
   *
   * @return string
   *   A string representing the contents of the DOM node.
   */
  protected function getInnerHtml(\DOMNode $node) {
    $html = '';
    foreach ($node->childNodes as $child) {
      $html .= $node->ownerDocument->saveHTML($child);
    }
    return $html;
  }

  /**
   * {@inheritdoc}
   */
  protected function getAttachmentData() {
    $data = $this->getFileData();
    return [
      'selector' => $this::SELECTOR ? $this::SELECTOR : $this->getFallBackSelector(),
      'attach' => $data['ATTACH'],
      'detach' => $data['DETACH'],
    ];
  }

  /**
   * {@inheritdoc}
   */
  protected function getCss() {
    return $this->getFileData()['CSS'];
  }

  /**
   * {@inheritdoc}
   */
  protected function getJs() {
    return $this->getFileData()['JS'];
  }

  /**
   * {@inheritdoc}
   */
  protected function getDependencies() {
    return $this->getFileData()['DEPENDENCIES'];
  }

  /**
   * {@inheritdoc}
   */
  protected function getTemplateData() {
    return $this->getFileData()['TEMPLATE'];
  }

  /**
   * {@inheritdoc}
   */
  protected function getComponentFileName() {
    return $this->pluginDefinition['simple_file'];
  }

  /**
   * {@inheritdoc}
   */
  protected function hasDependencies() {
    return (bool) $this->getFileData()['DEPENDENCIES'];
  }

  /**
   * {@inheritdoc}
   */
  protected function hasAttachments() {
    $data = $this->getFileData();
    return $data['ATTACH'] || $data['DETACH'];
  }

  /**
   * {@inheritdoc}
   */
  protected function hasCss() {
    return (bool) $this->getFileData()['CSS'];
  }

  /**
   * {@inheritdoc}
   */
  protected function hasJs() {
    return (bool) $this->getFileData()['JS'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildContextForm(array $form, FormStateInterface $form_state, array $default_values = []) {
    if ($callback = $this->getFileData()['buildContextForm']) {
      $form = call_user_func($callback, $form, $form_state, $default_values);
    }
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateContextForm(array &$form, FormStateInterface $form_state) {
    if ($callback = $this->getFileData()['validateContextForm']) {
      call_user_func($callback, $form, $form_state);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitContextForm(array &$form, FormStateInterface $form_state) {
    if ($callback = $this->getFileData()['submitContextForm']) {
      call_user_func($callback, $form, $form_state);
    }
  }

}
